@@ -1,3 +1,72 @@
+// RISC-V is a clean RISC ISA. It supports PC-relative load/store for
+// position-independent code. Its 32-bit and 64-bit ISAs are almost
+// identical. That is, you can think RV32 as a RV64 without 64-bit
+// operations. In this file, we support both RV64 and RV32.
+//
+// RISC-V is essentially little-endian, but the big-endian version is
+// available as an extension. GCC supports `-mbig-endian` to generate
+// big-endian code. Even in big-endian mode, machine instructions are
+// defined to be encoded in little-endian, though. Only the behavior of
+// load/store instructions are different between LE RISC-V and BE RISC-V.
+//
+// From the linker's point of view, the RISC-V's psABI is unique because
+// sections in input object files can be shrunk while being copied to the
+// output file. That is contrary to other psABIs in which sections are an
+// atomic unit of copying. Let me explain it in more details.
+//
+// Since RISC-V instructions are 16-bit or 32-bit long, there's no way to
+// embed a very large immediate into a branch instruction. In fact, JAL
+// (jump and link) instruction can jump to only within PC ± 1 MiB because
+// its immediate is only 21 bits long. If the destination is out of its
+// reach, we need to use two instructions instead; the first instruction
+// being AUIPC which sets upper 20 bits to a register and the second being
+// JALR with a 12-bit immediate and the register. Combined, they specify a
+// 32 bits displacement.
+//
+// Other RISC ISAs have the same limitation, and they solved the problem by
+// letting the linker create so-called "range extension thunks". It works as
+// follows: the compiler optimistically emits single jump instructions for
+// function calls. If the linker finds that a branch target is out of reach,
+// it emits a small piece of machine code near the branch instruction and
+// redirect the branch to the linker-synthesized code. The code constructs a
+// full 32-bit address in a register and jump to the destination. That
+// linker-synthesized code is called "range extension thunks" or just
+// "thunks".
+//
+// The RISC-V psABI is unique that it works the other way around. That is,
+// for RISC-V, the compiler always emits two instructions (AUIPC + JAL) for
+// function calls. If the linker finds the destination is reachable with a
+// single instruction, it replaces the two instructions with the one and
+// shrink the section size by one instruction length, instead of filling the
+// gap with a nop.
+//
+// With the presence of this relaxation, sections can no longer be
+// considered as an atomic unit. If we delete 4 bytes from the middle of a
+// section, all contents after that point needs to be shifted by 4. Symbol
+// values and relocation offsets have to be adjusted accordingly if they
+// refer past the deleted bytes.
+//
+// In mold, we use `r_deltas` to memorize how many bytes have be adjusted
+// for relocations. For symbols, we directly mutate their `value` member.
+//
+// RISC-V object files tend to have way more relocations than those for
+// other targets. This is because all branches, including ones that jump
+// within the same section, are explicitly expressed with relocations.
+// Here is why we need them: all control-flow statements such as `if` or
+// `for` are implemented using branch instructions. For other targets, the
+// compiler doesn't emit relocations for such branches because they know
+// at compile-time exactly how many bytes has to be skipped. That's not
+// true to RISC-V because the linker may delete bytes between a branch and
+// its destination. Therefore, all branches including in-section ones have
+// to be explicitly expressed with relocations.
+//
+// Note that this mechanism only shrink sections and never enlarge, as
+// the compiler always emits the longest instruction sequence. This
+// makes the linker implementation a bit simpler because we don't need
+// to worry about oscillation.
+//
+// https://github.com/riscv-non-isa/riscv-elf-psabi-doc/blob/master/riscv-elf.adoc
+
 #include "mold.h"
 
 #include <tbb/parallel_for.h>
@@ -22,7 +91,7 @@ static u32 utype(u32 val) {
   // U-type instructions are used in combination with I-type
   // instructions. U-type insn sets an immediate to the upper 20-bits
   // of a register. I-type insn sign-extends a 12-bits immediate and
-  // add it to a register value to construct a complete value. 0x800
+  // adds it to a register value to construct a complete value. 0x800
   // is added here to compensate for the sign-extension.
   return (val + 0x800) & 0xffff'f000;
 }
@@ -45,136 +114,126 @@ static u32 cjtype(u32 val) {
          bit(val, 1)  << 3  | bit(val, 5)  << 2;
 }
 
+static void write_itype(u8 *loc, u32 val) {
+  *(ul32 *)loc &= 0b000000'00000'11111'111'11111'1111111;
+  *(ul32 *)loc |= itype(val);
+}
+
+static void write_stype(u8 *loc, u32 val) {
+  *(ul32 *)loc &= 0b000000'11111'11111'111'00000'1111111;
+  *(ul32 *)loc |= stype(val);
+}
+
+static void write_btype(u8 *loc, u32 val) {
+  *(ul32 *)loc &= 0b000000'11111'11111'111'00000'1111111;
+  *(ul32 *)loc |= btype(val);
+}
+
+static void write_utype(u8 *loc, u32 val) {
+  *(ul32 *)loc &= 0b000000'00000'00000'000'11111'1111111;
+  *(ul32 *)loc |= utype(val);
+}
+
+static void write_jtype(u8 *loc, u32 val) {
+  *(ul32 *)loc &= 0b000000'00000'00000'000'11111'1111111;
+  *(ul32 *)loc |= jtype(val);
+}
+
+static void write_cbtype(u8 *loc, u32 val) {
+  *(ul16 *)loc &= 0b111'000'111'00000'11;
+  *(ul16 *)loc |= cbtype(val);
+}
+
+static void write_cjtype(u8 *loc, u32 val) {
+  *(ul16 *)loc &= 0b111'00000000000'11;
+  *(ul16 *)loc |= cjtype(val);
+}
+
+// Returns the rd register of an R/I/U/J-type instruction.
 static u32 get_rd(u32 val) {
   return bits(val, 11, 7);
 }
 
-static void write_itype(u8 *loc, u32 val) {
-  u32 mask = 0b000000'00000'11111'111'11111'1111111;
-  *(ul32 *)loc = (*(ul32 *)loc & mask) | itype(val);
-}
-
-static void write_stype(u8 *loc, u32 val) {
-  u32 mask = 0b000000'11111'11111'111'00000'1111111;
-  *(ul32 *)loc = (*(ul32 *)loc & mask) | stype(val);
-}
-
-static void write_btype(u8 *loc, u32 val) {
-  u32 mask = 0b000000'11111'11111'111'00000'1111111;
-  *(ul32 *)loc = (*(ul32 *)loc & mask) | btype(val);
-}
-
-static void write_utype(u8 *loc, u32 val) {
-  u32 mask = 0b000000'00000'00000'000'11111'1111111;
-  *(ul32 *)loc = (*(ul32 *)loc & mask) | utype(val);
-}
-
-static void write_jtype(u8 *loc, u32 val) {
-  u32 mask = 0b000000'00000'00000'000'11111'1111111;
-  *(ul32 *)loc = (*(ul32 *)loc & mask) | jtype(val);
-}
-
-static void write_cbtype(u8 *loc, u32 val) {
-  u32 mask = 0b1110001110000011;
-  *(ul16 *)loc = (*(ul16 *)loc & mask) | cbtype(val);
-}
-
-static void write_cjtype(u8 *loc, u32 val) {
-  u32 mask = 0b1110000000000011;
-  *(ul16 *)loc = (*(ul16 *)loc & mask) | cjtype(val);
+static void set_rs1(u8 *loc, u32 rs1) {
+  assert(rs1 < 32);
+  *(ul32 *)loc &= 0b111111'11111'00000'111'11111'1111111;
+  *(ul32 *)loc |= rs1 << 15;
 }
 
 template <typename E>
-static void write_plt_header(Context<E> &ctx) {
-  u8 *buf = ctx.buf + ctx.plt->shdr.sh_offset;
-
-  static const u32 plt0_64[] = {
-    0x00000397, // auipc  t2, %pcrel_hi(.got.plt)
-    0x41c30333, // sub    t1, t1, t3               # .plt entry + hdr + 12
-    0x0003be03, // ld     t3, %pcrel_lo(1b)(t2)    # _dl_runtime_resolve
-    0xfd430313, // addi   t1, t1, -44              # .plt entry
-    0x00038293, // addi   t0, t2, %pcrel_lo(1b)    # &.got.plt
-    0x00135313, // srli   t1, t1, 1                # .plt entry offset
-    0x0082b283, // ld     t0, 8(t0)                # link map
-    0x000e0067, // jr     t3
+void write_plt_header(Context<E> &ctx, u8 *buf) {
+  static const ul32 insn_64[] = {
+    0x0000'0397, // auipc  t2, %pcrel_hi(.got.plt)
+    0x41c3'0333, // sub    t1, t1, t3               # .plt entry + hdr + 12
+    0x0003'be03, // ld     t3, %pcrel_lo(1b)(t2)    # _dl_runtime_resolve
+    0xfd43'0313, // addi   t1, t1, -44              # .plt entry
+    0x0003'8293, // addi   t0, t2, %pcrel_lo(1b)    # &.got.plt
+    0x0013'5313, // srli   t1, t1, 1                # .plt entry offset
+    0x0082'b283, // ld     t0, 8(t0)                # link map
+    0x000e'0067, // jr     t3
   };
 
-  static const u32 plt0_32[] = {
-    0x00000397, // auipc  t2, %pcrel_hi(.got.plt)
-    0x41c30333, // sub    t1, t1, t3               # .plt entry + hdr + 12
-    0x0003ae03, // lw     t3, %pcrel_lo(1b)(t2)    # _dl_runtime_resolve
-    0xfd430313, // addi   t1, t1, -44              # .plt entry
-    0x00038293, // addi   t0, t2, %pcrel_lo(1b)    # &.got.plt
-    0x00235313, // srli   t1, t1, 2                # .plt entry offset
-    0x0042a283, // lw     t0, 4(t0)                # link map
-    0x000e0067, // jr     t3
+  static const ul32 insn_32[] = {
+    0x0000'0397, // auipc  t2, %pcrel_hi(.got.plt)
+    0x41c3'0333, // sub    t1, t1, t3               # .plt entry + hdr + 12
+    0x0003'ae03, // lw     t3, %pcrel_lo(1b)(t2)    # _dl_runtime_resolve
+    0xfd43'0313, // addi   t1, t1, -44              # .plt entry
+    0x0003'8293, // addi   t0, t2, %pcrel_lo(1b)    # &.got.plt
+    0x0023'5313, // srli   t1, t1, 2                # .plt entry offset
+    0x0042'a283, // lw     t0, 4(t0)                # link map
+    0x000e'0067, // jr     t3
   };
 
-  if constexpr (sizeof(Word<E>) == 8)
-    memcpy(buf, plt0_64, sizeof(plt0_64));
+  if constexpr (E::is_64)
+    memcpy(buf, insn_64, sizeof(insn_64));
   else
-    memcpy(buf, plt0_32, sizeof(plt0_32));
+    memcpy(buf, insn_32, sizeof(insn_32));
 
   u64 gotplt = ctx.gotplt->shdr.sh_addr;
   u64 plt = ctx.plt->shdr.sh_addr;
-
   write_utype(buf, gotplt - plt);
   write_itype(buf + 8, gotplt - plt);
   write_itype(buf + 16, gotplt - plt);
 }
 
-static constexpr u32 plt_entry_64[] = {
-  0x00000e17, // auipc   t3, %pcrel_hi(function@.got.plt)
-  0x000e3e03, // ld      t3, %pcrel_lo(1b)(t3)
-  0x000e0367, // jalr    t1, t3
-  0x00000013, // nop
+static const ul32 plt_entry_64[] = {
+  0x0000'0e17, // auipc   t3, %pcrel_hi(function@.got.plt)
+  0x000e'3e03, // ld      t3, %pcrel_lo(1b)(t3)
+  0x000e'0367, // jalr    t1, t3
+  0x0000'0013, // nop
 };
 
-static constexpr u32 plt_entry_32[] = {
-  0x00000e17, // auipc   t3, %pcrel_hi(function@.got.plt)
-  0x000e2e03, // lw      t3, %pcrel_lo(1b)(t3)
-  0x000e0367, // jalr    t1, t3
-  0x00000013, // nop
+static const ul32 plt_entry_32[] = {
+  0x0000'0e17, // auipc   t3, %pcrel_hi(function@.got.plt)
+  0x000e'2e03, // lw      t3, %pcrel_lo(1b)(t3)
+  0x000e'0367, // jalr    t1, t3
+  0x0000'0013, // nop
 };
 
 template <typename E>
-void PltSection<E>::copy_buf(Context<E> &ctx) {
-  u8 *buf = ctx.buf + ctx.plt->shdr.sh_offset;
+void write_plt_entry(Context<E> &ctx, u8 *buf, Symbol<E> &sym) {
+  if constexpr (E::is_64)
+    memcpy(buf, plt_entry_64, sizeof(plt_entry_64));
+  else
+    memcpy(buf, plt_entry_32, sizeof(plt_entry_32));
 
-  write_plt_header(ctx);
-
-  for (Symbol<E> *sym : symbols) {
-    u8 *ent = buf + E::plt_hdr_size + sym->get_plt_idx(ctx) * E::plt_size;
-    u64 gotplt = sym->get_gotplt_addr(ctx);
-    u64 plt = sym->get_plt_addr(ctx);
-
-    if constexpr (sizeof(Word<E>) == 8)
-      memcpy(ent, plt_entry_64, sizeof(plt_entry_64));
-    else
-      memcpy(ent, plt_entry_32, sizeof(plt_entry_32));
-
-    write_utype(ent, gotplt - plt);
-    write_itype(ent + 4, gotplt - plt);
-  }
+  u64 gotplt = sym.get_gotplt_addr(ctx);
+  u64 plt = sym.get_plt_addr(ctx);
+  write_utype(buf, gotplt - plt);
+  write_itype(buf + 4, gotplt - plt);
 }
 
 template <typename E>
-void PltGotSection<E>::copy_buf(Context<E> &ctx) {
-  u8 *buf = ctx.buf + this->shdr.sh_offset;
+void write_pltgot_entry(Context<E> &ctx, u8 *buf, Symbol<E> &sym) {
+  if constexpr (E::is_64)
+    memcpy(buf, plt_entry_64, sizeof(plt_entry_64));
+  else
+    memcpy(buf, plt_entry_32, sizeof(plt_entry_32));
 
-  for (Symbol<E> *sym : symbols) {
-    u8 *ent = buf + sym->get_pltgot_idx(ctx) * E::pltgot_size;
-    u64 got = sym->get_got_addr(ctx);
-    u64 plt = sym->get_plt_addr(ctx);
-
-    if constexpr (sizeof(Word<E>) == 8)
-      memcpy(ent, plt_entry_64, sizeof(plt_entry_64));
-    else
-      memcpy(ent, plt_entry_32, sizeof(plt_entry_32));
-
-    write_utype(ent, got - plt);
-    write_itype(ent + 4, got - plt);
-  }
+  u64 got = sym.get_got_addr(ctx);
+  u64 plt = sym.get_plt_addr(ctx);
+  write_utype(buf, got - plt);
+  write_itype(buf + 4, got - plt);
 }
 
 template <typename E>
@@ -183,121 +242,123 @@ void EhFrameSection<E>::apply_reloc(Context<E> &ctx, const ElfRel<E> &rel,
   u8 *loc = ctx.buf + this->shdr.sh_offset + offset;
 
   switch (rel.r_type) {
+  case R_NONE:
+    break;
   case R_RISCV_ADD32:
-    *(ul32 *)loc += val;
-    return;
+    *(U32<E> *)loc += val;
+    break;
   case R_RISCV_SUB8:
     *loc -= val;
-    return;
+    break;
   case R_RISCV_SUB16:
-    *(ul16 *)loc -= val;
-    return;
+    *(U16<E> *)loc -= val;
+    break;
   case R_RISCV_SUB32:
-    *(ul32 *)loc -= val;
-    return;
+    *(U32<E> *)loc -= val;
+    break;
   case R_RISCV_SUB6:
     *loc = (*loc & 0b1100'0000) | ((*loc - val) & 0b0011'1111);
-    return;
+    break;
   case R_RISCV_SET6:
     *loc = (*loc & 0b1100'0000) | (val & 0b0011'1111);
-    return;
+    break;
   case R_RISCV_SET8:
     *loc = val;
-    return;
+    break;
   case R_RISCV_SET16:
-    *(ul16 *)loc = val;
-    return;
+    *(U16<E> *)loc = val;
+    break;
   case R_RISCV_SET32:
-    *(ul32 *)loc = val;
-    return;
+    *(U32<E> *)loc = val;
+    break;
   case R_RISCV_32_PCREL:
-    *(ul32 *)loc = val - this->shdr.sh_addr - offset;
-    return;
+    *(U32<E> *)loc = val - this->shdr.sh_addr - offset;
+    break;
+  default:
+    Fatal(ctx) << "unsupported relocation in .eh_frame: " << rel;
   }
-  Fatal(ctx) << "unsupported relocation in .eh_frame: " << rel;
 }
 
 template <typename E>
 void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
-  ElfRel<E> *dynrel = nullptr;
   std::span<const ElfRel<E>> rels = get_rels(ctx);
 
-  i64 frag_idx = 0;
-
+  ElfRel<E> *dynrel = nullptr;
   if (ctx.reldyn)
     dynrel = (ElfRel<E> *)(ctx.buf + ctx.reldyn->shdr.sh_offset +
                            file.reldyn_offset + this->reldyn_offset);
 
+  auto get_r_delta = [&](i64 idx) {
+    return extra.r_deltas.empty() ? 0 : extra.r_deltas[idx];
+  };
+
   for (i64 i = 0; i < rels.size(); i++) {
     const ElfRel<E> &rel = rels[i];
-    if (rel.r_type == R_RISCV_NONE || rel.r_type == R_RISCV_RELAX)
+    if (rel.r_type == R_NONE || rel.r_type == R_RISCV_RELAX)
       continue;
 
     Symbol<E> &sym = *file.symbols[rel.r_sym];
-    i64 r_offset = rel.r_offset - extra.r_deltas[i];
-    i64 delta = extra.r_deltas[i + 1] - extra.r_deltas[i];
+    i64 r_offset = rel.r_offset - get_r_delta(i);
+    i64 removed_bytes = get_r_delta(i + 1) - get_r_delta(i);
     u8 *loc = base + r_offset;
 
-    const SectionFragmentRef<E> *frag_ref = nullptr;
-    if (rel_fragments && rel_fragments[frag_idx].idx == i)
-      frag_ref = &rel_fragments[frag_idx++];
-
-    auto overflow_check = [&](i64 val, i64 lo, i64 hi) {
+    auto check = [&](i64 val, i64 lo, i64 hi) {
       if (val < lo || hi <= val)
         Error(ctx) << *this << ": relocation " << rel << " against "
                    << sym << " out of range: " << val << " is not in ["
                    << lo << ", " << hi << ")";
     };
 
-#define S   (frag_ref ? frag_ref->frag->get_addr(ctx) : sym.get_addr(ctx))
-#define A   (frag_ref ? (u64)frag_ref->addend : (u64)rel.r_addend)
+#define S   sym.get_addr(ctx)
+#define A   rel.r_addend
 #define P   (get_addr() + r_offset)
 #define G   (sym.get_got_idx(ctx) * sizeof(Word<E>))
 #define GOT ctx.got->shdr.sh_addr
 
     switch (rel.r_type) {
     case R_RISCV_32:
-      if constexpr (sizeof(Word<E>) == 4)
-        apply_abs_dyn_rel(ctx, sym, rel, loc, S, A, P, dynrel);
+      if constexpr (E::is_64)
+        *(U32<E> *)loc = S + A;
       else
-        *(ul32 *)loc = S + A;
+        apply_dyn_absrel(ctx, sym, rel, loc, S, A, P, dynrel);
       break;
     case R_RISCV_64:
-      assert(sizeof(Word<E>) == 8);
-      apply_abs_dyn_rel(ctx, sym, rel, loc, S, A, P, dynrel);
+      assert(E::is_64);
+      apply_dyn_absrel(ctx, sym, rel, loc, S, A, P, dynrel);
       break;
     case R_RISCV_BRANCH: {
       i64 val = S + A - P;
-      overflow_check(val, -(1 << 12), 1 << 12);
+      check(val, -(1 << 12), 1 << 12);
       write_btype(loc, val);
       break;
     }
     case R_RISCV_JAL: {
       i64 val = S + A - P;
-      overflow_check(val, -(1 << 20), 1 << 20);
+      check(val, -(1 << 20), 1 << 20);
       write_jtype(loc, val);
       break;
     }
     case R_RISCV_CALL:
     case R_RISCV_CALL_PLT: {
       u32 rd = get_rd(*(ul32 *)(contents.data() + rel.r_offset + 4));
-      if (delta == 4) {
+
+      if (removed_bytes == 4) {
         // auipc + jalr -> jal
         *(ul32 *)loc = (rd << 7) | 0b1101111;
         write_jtype(loc, S + A - P);
-      } else if (delta == 6 && rd == 0) {
+      } else if (removed_bytes == 6 && rd == 0) {
         // auipc + jalr -> c.j
         *(ul16 *)loc = 0b101'00000000000'01;
         write_cjtype(loc, S + A - P);
-      } else if (delta == 6 && rd == 1) {
+      } else if (removed_bytes == 6 && rd == 1) {
         // auipc + jalr -> c.jal
-        assert(sizeof(Word<E>) == 4);
+        assert(!E::is_64);
         *(ul16 *)loc = 0b001'00000000000'01;
         write_cjtype(loc, S + A - P);
       } else {
-        assert(delta == 0);
+        assert(removed_bytes == 0);
         u64 val = sym.esym().is_undef_weak() ? 0 : S + A - P;
-        overflow_check(val, -(1LL << 31), 1LL << 31);
+        check(val, -(1LL << 31), 1LL << 31);
         write_utype(loc, val);
         write_itype(loc + 4, val);
       }
@@ -322,28 +383,42 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
         *(ul32 *)loc = S + A - P;
       }
       break;
-    case R_RISCV_LO12_I:
-      write_itype(loc, S + A);
-      break;
-    case R_RISCV_LO12_S:
-      write_stype(loc, S + A);
-      break;
     case R_RISCV_HI20: {
       i64 val = S + A;
-      overflow_check(val, -(1LL << 31), 1LL << 31);
-      write_utype(loc, val);
+      if (removed_bytes == 0) {
+        check(val, -(1LL << 31), 1LL << 31);
+        write_utype(loc, val);
+      } else {
+        assert(removed_bytes == 4);
+        assert(sign_extend(val, 11) == val);
+      }
+      break;
+    }
+    case R_RISCV_LO12_I:
+    case R_RISCV_LO12_S: {
+      i64 val = S + A;
+      if (rel.r_type == R_RISCV_LO12_I)
+        write_itype(loc, val);
+      else
+        write_stype(loc, val);
+
+      // Rewrite `lw t1, 0(t0)` with `lw t1, 0(x0)` if the address is
+      // accessible relative to the zero register. If the upper 20 bits
+      // are all zero, the corresponding LUI might have been removed.
+      if (sign_extend(val, 11) == val)
+        set_rs1(loc, 0);
       break;
     }
     case R_RISCV_TPREL_HI20:
-      assert(delta == 0 || delta == 4);
-      if (delta == 0)
-        write_utype(loc, S + A - ctx.tls_begin);
+      assert(removed_bytes == 0 || removed_bytes == 4);
+      if (removed_bytes == 0)
+        write_utype(loc, S + A - ctx.tp_addr);
       break;
     case R_RISCV_TPREL_ADD:
       break;
     case R_RISCV_TPREL_LO12_I:
     case R_RISCV_TPREL_LO12_S: {
-      i64 val = S + A - ctx.tls_begin;
+      i64 val = S + A - ctx.tp_addr;
       if (rel.r_type == R_RISCV_TPREL_LO12_I)
         write_itype(loc, val);
       else
@@ -352,33 +427,32 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
       // Rewrite `lw t1, 0(t0)` with `lw t1, 0(tp)` if the address is
       // directly accessible using tp. tp is x4.
       if (sign_extend(val, 11) == val)
-        *(ul32 *)loc = (*(ul32 *)loc & 0b111111'11111'00000'111'11111'1111111) |
-                       (4 << 15);
+        set_rs1(loc, 4);
       break;
     }
     case R_RISCV_ADD8:
       loc += S + A;
       break;
     case R_RISCV_ADD16:
-      *(ul16 *)loc += S + A;
+      *(U16<E> *)loc += S + A;
       break;
     case R_RISCV_ADD32:
-      *(ul32 *)loc += S + A;
+      *(U32<E> *)loc += S + A;
       break;
     case R_RISCV_ADD64:
-      *(ul64 *)loc += S + A;
+      *(U64<E> *)loc += S + A;
       break;
     case R_RISCV_SUB8:
       loc -= S + A;
       break;
     case R_RISCV_SUB16:
-      *(ul16 *)loc -= S + A;
+      *(U16<E> *)loc -= S + A;
       break;
     case R_RISCV_SUB32:
-      *(ul32 *)loc -= S + A;
+      *(U32<E> *)loc -= S + A;
       break;
     case R_RISCV_SUB64:
-      *(ul64 *)loc -= S + A;
+      *(U64<E> *)loc -= S + A;
       break;
     case R_RISCV_ALIGN: {
       // A R_RISCV_ALIGN is followed by a NOP sequence. We need to remove
@@ -386,27 +460,27 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
       // aligned to a given alignment boundary.
       //
       // We need to guarantee that the NOP sequence is valid after byte
-      // removal (e.g. we can't remove 2 bytes from a 4-byte NOP). For the
-      // sake of simplicity, we always rewrite the entire NOP sequence.
+      // removal (e.g. we can't remove the first 2 bytes of a 4-byte NOP).
+      // For the sake of simplicity, we always rewrite the entire NOP sequence.
       i64 padding_size = align_to(P, bit_ceil(rel.r_addend + 1)) - P;
       assert(padding_size % 2 == 0);
 
       i64 i = 0;
       for (; i <= padding_size - 4; i += 4)
-        *(ul32 *)(loc + i) = 0x00000013; // nop
+        *(ul32 *)(loc + i) = 0x0000'0013; // nop
       if (i != padding_size)
         *(ul16 *)(loc + i) = 0x0001;     // c.nop
       break;
     }
     case R_RISCV_RVC_BRANCH: {
       i64 val = S + A - P;
-      overflow_check(val, -(1 << 8), 1 << 8);
+      check(val, -(1 << 8), 1 << 8);
       write_cbtype(loc, val);
       break;
     }
     case R_RISCV_RVC_JUMP: {
       i64 val = S + A - P;
-      overflow_check(val, -(1 << 11), 1 << 11);
+      check(val, -(1 << 11), 1 << 11);
       write_cjtype(loc, val);
       break;
     }
@@ -420,13 +494,13 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
       *loc = S + A;
       break;
     case R_RISCV_SET16:
-      *(ul16 *)loc = S + A;
+      *(U16<E> *)loc = S + A;
       break;
     case R_RISCV_SET32:
-      *(ul32 *)loc = S + A;
+      *(U32<E> *)loc = S + A;
       break;
     case R_RISCV_32_PCREL:
-      *(ul32 *)loc = S + A - P;
+      *(U32<E> *)loc = S + A - P;
       break;
     case R_RISCV_PCREL_LO12_I:
     case R_RISCV_PCREL_LO12_S:
@@ -443,9 +517,9 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
 #undef GOT
   }
 
-  // Handle LO12 relocations. In the above loop, PC-relative HI20
+  // Handle PC-relative LO12 relocations. In the above loop, pcrel HI20
   // relocations overwrote instructions with full 32-bit values to allow
-  // their corresponding LO12 relocations to read their values.
+  // their corresponding pcrel LO12 relocations to read their values.
   for (i64 i = 0; i < rels.size(); i++) {
     switch (rels[i].r_type)
     case R_RISCV_PCREL_LO12_I:
@@ -453,7 +527,7 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
       Symbol<E> &sym = *file.symbols[rels[i].r_sym];
       assert(sym.get_input_section() == this);
 
-      u8 *loc = base + rels[i].r_offset - extra.r_deltas[i];
+      u8 *loc = base + rels[i].r_offset - get_r_delta(i);
       u32 val = *(ul32 *)(base + sym.value);
 
       if (rels[i].r_type == R_RISCV_PCREL_LO12_I)
@@ -463,16 +537,16 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
     }
   }
 
-  // Restore the original instructions HI20 relocations overwrote.
+  // Restore the original instructions pcrel HI20 relocations overwrote.
   for (i64 i = 0; i < rels.size(); i++) {
     switch (rels[i].r_type) {
     case R_RISCV_GOT_HI20:
     case R_RISCV_PCREL_HI20:
     case R_RISCV_TLS_GOT_HI20:
     case R_RISCV_TLS_GD_HI20: {
-      u8 *loc = base + rels[i].r_offset - extra.r_deltas[i];
+      u8 *loc = base + rels[i].r_offset - get_r_delta(i);
       u32 val = *(ul32 *)loc;
-      *(ul32 *)loc = *(ul32 *)(contents.data() + rels[i].r_offset);
+      memcpy(loc, contents.data() + rels[i].r_offset, 4);
       write_utype(loc, val);
     }
     }
@@ -485,7 +559,7 @@ void InputSection<E>::apply_reloc_nonalloc(Context<E> &ctx, u8 *base) {
 
   for (i64 i = 0; i < rels.size(); i++) {
     const ElfRel<E> &rel = rels[i];
-    if (rel.r_type == R_RISCV_NONE)
+    if (rel.r_type == R_NONE)
       continue;
 
     Symbol<E> &sym = *file.symbols[rel.r_sym];
@@ -497,48 +571,45 @@ void InputSection<E>::apply_reloc_nonalloc(Context<E> &ctx, u8 *base) {
     }
 
     SectionFragment<E> *frag;
-    i64 addend;
-    std::tie(frag, addend) = get_fragment(ctx, rel);
+    i64 frag_addend;
+    std::tie(frag, frag_addend) = get_fragment(ctx, rel);
 
 #define S (frag ? frag->get_addr(ctx) : sym.get_addr(ctx))
-#define A (frag ? (u64)addend : (u64)rel.r_addend)
+#define A (frag ? frag_addend : (i64)rel.r_addend)
 
     switch (rel.r_type) {
     case R_RISCV_32:
-      *(ul32 *)loc = S + A;
+      *(U32<E> *)loc = S + A;
       break;
     case R_RISCV_64:
-      if (!frag) {
-        if (std::optional<u64> val = get_tombstone(sym)) {
-          *(ul64 *)loc = *val;
-          break;
-        }
-      }
-      *(ul64 *)loc = S + A;
+      if (std::optional<u64> val = get_tombstone(sym, frag))
+        *(U64<E> *)loc = *val;
+      else
+        *(U64<E> *)loc = S + A;
       break;
     case R_RISCV_ADD8:
       *loc += S + A;
       break;
     case R_RISCV_ADD16:
-      *(ul16 *)loc += S + A;
+      *(U16<E> *)loc += S + A;
       break;
     case R_RISCV_ADD32:
-      *(ul32 *)loc += S + A;
+      *(U32<E> *)loc += S + A;
       break;
     case R_RISCV_ADD64:
-      *(ul64 *)loc += S + A;
+      *(U64<E> *)loc += S + A;
       break;
     case R_RISCV_SUB8:
       *loc -= S + A;
       break;
     case R_RISCV_SUB16:
-      *(ul16 *)loc -= S + A;
+      *(U16<E> *)loc -= S + A;
       break;
     case R_RISCV_SUB32:
-      *(ul32 *)loc -= S + A;
+      *(U32<E> *)loc -= S + A;
       break;
     case R_RISCV_SUB64:
-      *(ul64 *)loc -= S + A;
+      *(U64<E> *)loc -= S + A;
       break;
     case R_RISCV_SUB6:
       *loc = (*loc & 0b1100'0000) | ((*loc - (S + A)) & 0b0011'1111);
@@ -550,10 +621,10 @@ void InputSection<E>::apply_reloc_nonalloc(Context<E> &ctx, u8 *base) {
       *loc = S + A;
       break;
     case R_RISCV_SET16:
-      *(ul16 *)loc = S + A;
+      *(U16<E> *)loc = S + A;
       break;
     case R_RISCV_SET32:
-      *(ul32 *)loc = S + A;
+      *(U32<E> *)loc = S + A;
       break;
     default:
       Fatal(ctx) << *this << ": invalid relocation for non-allocated sections: "
@@ -568,12 +639,9 @@ void InputSection<E>::apply_reloc_nonalloc(Context<E> &ctx, u8 *base) {
 
 template <typename E>
 void InputSection<E>::copy_contents_riscv(Context<E> &ctx, u8 *buf) {
-  // A non-alloc section isn't relaxed, so just copy it as one big chunk.
-  if (!(shdr().sh_flags & SHF_ALLOC)) {
-    if (compressed)
-      uncompress_to(ctx, buf);
-    else
-      memcpy(buf, contents.data(), contents.size());
+  // If a section is not relaxed, we can copy it as a one big chunk.
+  if (extra.r_deltas.empty()) {
+    uncompress_to(ctx, buf);
     return;
   }
 
@@ -607,7 +675,7 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
   // Scan relocations
   for (i64 i = 0; i < rels.size(); i++) {
     const ElfRel<E> &rel = rels[i];
-    if (rel.r_type == R_RISCV_NONE)
+    if (rel.r_type == R_NONE)
       continue;
 
     Symbol<E> &sym = *file.symbols[rel.r_sym];
@@ -617,23 +685,23 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
       continue;
     }
 
-    if (sym.get_type() == STT_GNU_IFUNC)
+    if (sym.is_ifunc())
       sym.flags |= (NEEDS_GOT | NEEDS_PLT);
 
     switch (rel.r_type) {
     case R_RISCV_32:
-      if constexpr (sizeof(Word<E>) == 8)
-        scan_abs_rel(ctx, sym, rel);
+      if constexpr (E::is_64)
+        scan_rel(ctx, sym, rel, absrel_table);
       else
-        scan_abs_dyn_rel(ctx, sym, rel);
+        scan_rel(ctx, sym, rel, dyn_absrel_table);
       break;
     case R_RISCV_HI20:
-      scan_abs_rel(ctx, sym, rel);
+      scan_rel(ctx, sym, rel, absrel_table);
       break;
     case R_RISCV_64:
-      if constexpr (sizeof(Word<E>) == 4)
+      if constexpr (!E::is_64)
         Fatal(ctx) << *this << ": R_RISCV_64 cannot be used on RV32";
-      scan_abs_dyn_rel(ctx, sym, rel);
+      scan_rel(ctx, sym, rel, dyn_absrel_table);
       break;
     case R_RISCV_CALL:
     case R_RISCV_CALL_PLT:
@@ -651,7 +719,7 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
       sym.flags |= NEEDS_TLSGD;
       break;
     case R_RISCV_32_PCREL:
-      scan_pcrel_rel(ctx, sym, rel);
+      scan_rel(ctx, sym, rel, pcrel_table);
       break;
     case R_RISCV_BRANCH:
     case R_RISCV_JAL:
@@ -690,17 +758,8 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
 
 template <typename E>
 static bool is_resizable(Context<E> &ctx, InputSection<E> *isec) {
-  return isec && (isec->shdr().sh_flags & SHF_ALLOC);
-}
-
-template <typename E>
-static std::vector<Symbol<E> *> get_sorted_symbols(InputSection<E> &isec) {
-  std::vector<Symbol<E> *> vec;
-  for (Symbol<E> *sym : isec.file.symbols)
-    if (sym->file == &isec.file && sym->get_input_section() == &isec)
-      vec.push_back(sym);
-  sort(vec, [](Symbol<E> *a, Symbol<E> *b) { return a->value < b->value; });
-  return vec;
+  return isec && isec->is_alive && (isec->shdr().sh_flags & SHF_ALLOC) &&
+         (isec->shdr().sh_flags & SHF_EXECINSTR);
 }
 
 // Returns the distance between a relocated place and a symbol.
@@ -759,10 +818,18 @@ static void shrink_section(Context<E> &ctx, InputSection<E> &isec, bool use_rvc)
         rels[i + 1].r_type != R_RISCV_RELAX)
       continue;
 
+    // Linker-synthesized symbols haven't been assigned their final
+    // values when we are shrinking sections because actual values can
+    // be computed only after we fix the file layout. Therefore, we
+    // assume that relocations against such symbols are always
+    // non-relaxable.
+    if (sym.file == ctx.internal_obj)
+      continue;
+
     switch (r.r_type) {
     case R_RISCV_CALL:
     case R_RISCV_CALL_PLT: {
-      // These relocations refer  an AUIPC + JALR instruction pair to
+      // These relocations refer an AUIPC + JALR instruction pair to
       // allow to jump to anywhere in PC ± 2 GiB. If the jump target is
       // close enough to PC, we can use C.J, C.JAL or JAL instead.
       i64 dist = compute_distance(ctx, sym, isec, r);
@@ -776,8 +843,7 @@ static void shrink_section(Context<E> &ctx, InputSection<E> &isec, bool use_rvc)
         // If rd is x0 and the jump target is within ±2 KiB, we can use
         // C.J, saving 6 bytes.
         delta += 6;
-      } else if (rd == 1 && sign_extend(dist, 11) == dist
-                 && use_rvc && sizeof(Word<E>) == 4) {
+      } else if (rd == 1 && sign_extend(dist, 11) == dist && use_rvc && !E::is_64) {
         // If rd is x1 and the jump target is within ±2 KiB, we can use
         // C.JAL. This is RV32 only because C.JAL is RV32-only instruction.
         delta += 6;
@@ -787,6 +853,15 @@ static void shrink_section(Context<E> &ctx, InputSection<E> &isec, bool use_rvc)
       }
       break;
     }
+    case R_RISCV_HI20: {
+      // If the upper 20 bits are all zero, we can remove LUI.
+      // The corresponding instructions referred by LO12_I/LO12_S
+      // relocations will use the zero register instead.
+      i64 val = sym.get_addr(ctx);
+      if (sign_extend(val, 11) == val)
+        delta += 4;
+      break;
+    }
     case R_RISCV_TPREL_HI20:
     case R_RISCV_TPREL_ADD: {
       // These relocations are used to materialize the upper 20 bits of
@@ -807,7 +882,7 @@ static void shrink_section(Context<E> &ctx, InputSection<E> &isec, bool use_rvc)
       //  sw   t0,%tprel_lo(foo)(tp)
       //
       // Here, we remove `lui` and `add` if the offset is within ±2 KiB.
-      i64 val = sym.get_addr(ctx) + r.r_addend - ctx.tls_begin;
+      i64 val = sym.get_addr(ctx) + r.r_addend - ctx.tp_addr;
       if (sign_extend(val, 11) == val)
         delta += 4;
       break;
@@ -819,68 +894,23 @@ static void shrink_section(Context<E> &ctx, InputSection<E> &isec, bool use_rvc)
   isec.sh_size -= delta;
 }
 
-template <typename E>
-static void update_symbol_values(Context<E> &ctx, InputSection<E> &isec) {
-  std::span<const ElfRel<E>> rels = isec.get_rels(ctx);
-  i64 i = 0;
-  for (Symbol<E> *sym : get_sorted_symbols(isec)) {
-    while (i < rels.size() && rels[i].r_offset < sym->value)
-      i++;
-    sym->value -= isec.extra.r_deltas[i];
-  }
-}
-
-// RISC-V instructions are 16 or 32 bits long, so immediates encoded
-// in instructions can't be 32 bits long. Therefore, branch and load
-// instructions can't refer the 4 GiB address space unlike x86-64.
-// In fact, JAL (jump and link) instruction can jump to only within
-// ±1 MiB as their immediate is only 21 bits long.
+// Shrink sections by interpreting relocations.
 //
-// If you want to jump to somewhere further than that, you need to
-// construct a full 32-bit offset using multiple instruction and
-// branch to that place (e.g. AUIPC and JALR instead of JAL).
-// In this comment, we refer instructions such as JAL as the short
-// encoding and ones such as AUIPC+JALR as the long encoding.
-//
-// By default, compiler always uses the long encoding so that branch
-// targets are always encodable. This is a safe bet for them but
-// may result in inefficient code. Therefore, the RISC-V psABI defines
-// a mechanism for the linker to replace long encoding instructions
-// with short ones, shrinking the section and increasing the code
-// density.
-//
-// This is contrary to the psABIs for the other RISC processors such as
-// ARM64. Typically, they use short instructions by default, and a
-// linker creates so-called "thunks" to extend ranges of short jumps.
-// On RISC-V, instructions are in the long encoding by default, and
-// the linker shrinks them if it can.
-//
-// When we shrink a section, we need to adjust relocation offsets and
-// symbol values. For example, if we replace AUIPC+JALR with JAL
-// (which saves 4 bytes), all relocations pointing to anywhere after
-// that location need to be shifted by 4. In addition to that, any
-// symbol that refers anywhere after that locatioin need to be shifted
-// by 4 bytes as well.
-//
-// For relocations, we use `r_deltas` array to memorize how many bytes
-// have be adjusted. For symbols, we directly mutate their `value`
-// member.
-//
-// This operation seems to be optional, as by default instructions are
-// using the long encoding, but calling this function is actually
-// mandatory because of R_RISCV_ALIGN. R_RISCV_ALIGN relocation is a
-// directive to the linker to align the location referred to by the
-// relocation to a specified byte boundary. We at least have to
-// interpret them satisfy the constraints imposed by R_RISCV_ALIGN
-// relocations.
+// This operation seems to be optional, because by default longest
+// instructions are being used. However, calling this function is actually
+// mandatory because of R_RISCV_ALIGN. R_RISCV_ALIGN is a directive to the
+// linker to align the location referred to by the relocation to a
+// specified byte boundary. We at least have to interpret them to satisfy
+// the alignment constraints.
 template <typename E>
 i64 riscv_resize_sections(Context<E> &ctx) {
   Timer t(ctx, "riscv_resize_sections");
 
-  // True if we can use the 2-byte instructions.
+  // True if we can use the 2-byte instructions. This is usually true on
+  // Unix because RV64GC is generally considered the baseline hardware.
   bool use_rvc = get_eflags(ctx) & EF_RISCV_RVC;
 
-  // Find R_RISCV_CALL AND R_RISCV_CALL_PLT that can be relaxed.
+  // Find all the relocations that can be relaxed.
   // This step should only shrink sections.
   tbb::parallel_for_each(ctx.objs, [&](ObjectFile<E> *file) {
     for (std::unique_ptr<InputSection<E>> &isec : file->sections)
@@ -888,10 +918,24 @@ i64 riscv_resize_sections(Context<E> &ctx) {
         shrink_section(ctx, *isec, use_rvc);
   });
 
+  // Fix symbol values.
   tbb::parallel_for_each(ctx.objs, [&](ObjectFile<E> *file) {
-    for (std::unique_ptr<InputSection<E>> &isec : file->sections)
-      if (is_resizable(ctx, isec.get()))
-        update_symbol_values(ctx, *isec);
+    for (Symbol<E> *sym : file->symbols) {
+      if (sym->file != file)
+        continue;
+
+      InputSection<E> *isec = sym->get_input_section();
+      if (!isec || isec->extra.r_deltas.empty())
+        continue;
+
+      std::span<const ElfRel<E>> rels = isec->get_rels(ctx);
+      auto it = std::lower_bound(rels.begin(), rels.end(), sym->value,
+                                 [&](const ElfRel<E> &r, u64 val) {
+        return r.r_offset < val;
+      });
+
+      sym->value -= isec->extra.r_deltas[it - rels.begin()];
+    }
   });
 
   // Re-compute section offset again to finalize them.
@@ -899,9 +943,10 @@ i64 riscv_resize_sections(Context<E> &ctx) {
   return set_osec_offsets(ctx);
 }
 
-#define INSTANTIATE_RISCV(E)                                                 \
-  template void PltSection<E>::copy_buf(Context<E> &);                       \
-  template void PltGotSection<E>::copy_buf(Context<E> &);                    \
+#define INSTANTIATE(E)                                                       \
+  template void write_plt_header(Context<E> &, u8 *);                        \
+  template void write_plt_entry(Context<E> &, u8 *, Symbol<E> &);            \
+  template void write_pltgot_entry(Context<E> &, u8 *, Symbol<E> &);         \
   template void                                                              \
   EhFrameSection<E>::apply_reloc(Context<E> &, const ElfRel<E> &, u64, u64); \
   template void InputSection<E>::apply_reloc_alloc(Context<E> &, u8 *);      \
@@ -910,7 +955,9 @@ i64 riscv_resize_sections(Context<E> &ctx) {
   template void InputSection<E>::scan_relocations(Context<E> &);             \
   template i64 riscv_resize_sections(Context<E> &);
 
-INSTANTIATE_RISCV(RISCV64);
-INSTANTIATE_RISCV(RISCV32);
+INSTANTIATE(RV64LE);
+INSTANTIATE(RV64BE);
+INSTANTIATE(RV32LE);
+INSTANTIATE(RV32BE);
 
 } // namespace mold::elf
