@@ -1,6 +1,7 @@
 #include "mold.h"
 
 #include <cstring>
+#include <filesystem>
 #include <signal.h>
 #include <tbb/global_control.h>
 
@@ -8,8 +9,15 @@
 #include <mimalloc-new-delete.h>
 #endif
 
+#ifdef __FreeBSD__
+# include <sys/sysctl.h>
+# include <unistd.h>
+#endif
+
 namespace mold {
 
+std::string mold_version_string = MOLD_VERSION;
+
 namespace elf {
 int main(int argc, char **argv);
 }
@@ -27,8 +35,36 @@ static std::string get_mold_version() {
 void cleanup() {
   if (output_tmpfile)
     unlink(output_tmpfile);
-  if (socket_tmpfile)
-    unlink(socket_tmpfile);
+}
+
+std::string errno_string() {
+  // strerror is not thread-safe, so guard it with a lock.
+  static std::mutex mu;
+  std::scoped_lock lock(mu);
+  return strerror(errno);
+}
+
+// Returns the path of the mold executable itself
+std::string get_self_path() {
+#ifdef __FreeBSD__
+  // /proc may not be mounted on FreeBSD. The proper way to get the
+  // current executable's path is to use sysctl(2).
+  int mib[4];
+  mib[0] = CTL_KERN;
+  mib[1] = KERN_PROC;
+  mib[2] = KERN_PROC_PATHNAME;
+  mib[3] = -1;
+
+  size_t size;
+  sysctl(mib, 4, NULL, &size, NULL, 0);
+
+  std::string path;
+  path.resize(size);
+  sysctl(mib, 4, path.data(), &size, NULL, 0);
+  return path;
+#else
+  return std::filesystem::read_symlink("/proc/self/exe").string();
+#endif
 }
 
 // mold mmap's an output file, and the mmap succeeds even if there's
