@@ -1,16 +1,5 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 cat <<EOF | $CC -c -o $t/a.o -xc -
 #include <stdio.h>
@@ -52,5 +41,3 @@ $QEMU $t/exe | grep -q '^wrap_foo$'
 
 $CC -B. -o $t/exe $t/a.o $t/c.o -Wl,-wrap,foo
 $QEMU $t/exe | grep -q '^foo$'
-
-echo OK
