@@ -1,15 +1,9 @@
 #!/bin/bash
 export LC_ALL=C
 set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
 testname=$(basename "$0" .sh)
 echo -n "Testing $testname ... "
-t=out/test/macho/$MACHINE/$testname
+t=out/test/macho/$(uname -m)/$testname
 mkdir -p $t
 
 mkdir -p $t/libs/SomeFramework.framework/
@@ -36,7 +30,7 @@ exports:
 ...
 EOF
 
-cat <<EOF | clang -o $t/a.o -c -xc -
+cat <<EOF | cc -o $t/a.o -c -xc -
 extern void foo();
 extern void bar() __attribute__((weak_import));
 
@@ -47,7 +41,7 @@ int main() {
 }
 EOF
 
-clang --ld-path=./ld64 -o $t/exe $t/a.o -F$t/libs/ -Wl,-framework,SomeFramework
+cc --ld-path=./ld64 -o $t/exe $t/a.o -F$t/libs/ -Wl,-framework,SomeFramework
 
 otool -L $t/exe | grep -q '/usr/frameworks/SomeFramework.framework/SomeFramework'
 
