@@ -1,21 +1,15 @@
 #!/bin/bash
 export LC_ALL=C
 set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
 testname=$(basename "$0" .sh)
 echo -n "Testing $testname ... "
-t=out/test/macho/$MACHINE/$testname
+t=out/test/macho/$(uname -m)/$testname
 mkdir -p $t
 
 # For some reason, this test fails only on GitHub CI.
 [ "$GITHUB_ACTIONS" = true ] && { echo skipped; exit; }
 
-cat <<EOF | $CC -o $t/a.o -c -xc -
+cat <<EOF | cc -o $t/a.o -c -xc -
 #include <stdio.h>
 
 _Thread_local int a;
@@ -29,7 +23,7 @@ int main() {
 }
 EOF
 
-clang --ld-path=./ld64 -o $t/exe $t/a.o
+cc --ld-path=./ld64 -o $t/exe $t/a.o
 $t/exe | grep -q '^0 5 5$'
 
 echo OK
